from astroquery.gaia import Gaia
from astroquery.utils.tap.core import TapPlus
from astropy.table import Table, Column
import numpy as np

# 2024-04-19: Gaia archive down.
# let us query our Gaia stars from Vizier instead.
vizier_to_gaia_column_mapping = {
    'RA_ICRS': 'ra',
    'DE_ICRS': 'dec',
    'Gmag': 'phot_g_mean_mag',
    'RPmag': 'phot_rp_mean_mag',
    'BPmag': 'phot_bp_mean_mag',
    'pmRA': 'pmra',
    'pmDE': 'pmdec',
    'Source': 'source_id',
    'sepsi': 'astrometric_excess_noise_sig',
    'RFG': 'phot_g_mean_flux_over_error'
}
gaia_to_vizier_column_mapping = {value: key for key, value in vizier_to_gaia_column_mapping.items()}


def construct_where_conditions(gaia_provider, astrometric_excess_noise_max=None, gmag_range=None,
                               min_phot_g_mean_flux_over_error=None):
    """
    utility function for the find functions below, just checks on the conditions and returns a list of potential
    statements to us after 'WHERE'. Also takes care of formatting the table from which we will query.
    Args:
        gaia_provider: string, 'gaia' or 'vizier'
        astrometric_excess_noise_max: float, default None
        gmag_range: tuple of floats (min, max), default None
        min_phot_g_mean_flux_over_error: float, default None

    Returns:
        list of strings containing the conditions, and string containing the name of the table to query.

    """
    gaia_provider = gaia_provider.lower()
    assert gaia_provider in ['gaia', 'vizier'], "gaia_provider must be either 'gaia' or 'vizier'"
    if gaia_provider == 'gaia':
        query_table = "gaiadr3.gaia_source as gdr3 "
    else:
        query_table = f'"I/355/gaiadr3" AS gdr3 '

    where_conditions = []

    if astrometric_excess_noise_max is not None:
        col_name = 'astrometric_excess_noise_sig'
        if gaia_provider == 'vizier':
            col_name = gaia_to_vizier_column_mapping[col_name]
        where_conditions.append(f"gdr3.{col_name} < {astrometric_excess_noise_max}")

    if gmag_range is not None:
        col_name = 'phot_g_mean_mag'
        if gaia_provider == 'vizier':
            col_name = gaia_to_vizier_column_mapping[col_name]
        where_conditions.append(f"gdr3.{col_name} BETWEEN {gmag_range[0]} AND {gmag_range[1]}")

    if min_phot_g_mean_flux_over_error is not None:
        col_name = 'phot_g_mean_flux_over_error'
        if gaia_provider == 'vizier':
            col_name = gaia_to_vizier_column_mapping[col_name]
        where_conditions.append(f"gdr3.{col_name} > {min_phot_g_mean_flux_over_error}")

    return where_conditions, query_table


def find_gaia_stars(region_type, *args, **kwargs):
    """
    Main function to query Gaia stars based on region type (Circle or Polygon).

    :param region_type: str, 'circle' or 'polygon' to define the type of region for the query.
    :param args: Arguments passed to the specific region query function.
    :param kwargs: Keyword arguments for filtering options, passed to the specific region query function.
    """
    if region_type.lower() == 'circle':
        stars_table = find_gaia_stars_in_circle(*args, **kwargs)
    elif region_type.lower() == 'polygon':
        stars_table = find_gaia_stars_in_polygon(*args, **kwargs)
    else:
        raise ValueError("region_type must be either 'Circle' or 'Polygon'")

    # it seems that there is some variation when querying gaia? sometimes columns are capitalized, others not.
    # so, force lower:
    for name in stars_table.colnames:
        new_name = name.lower()
        stars_table.rename_column(name, new_name)
    return stars_table


def run_query(gaia_provider, adql_query):
    """
    Utility function to run the adql query constructed by other functions, given a gaia provider (gaia or vizier)
    Args:
        gaia_provider: str, 'gaia' or 'vizier'
        adql_query: str, an adql query.

    Returns:
        astropy table of sources, with columns following the gaia archive labelling.
    """
    if gaia_provider.lower() == 'gaia':
        Gaia.MAIN_GAIA_TABLE = 'gaiadr3.gaia_source'
        Gaia.ROW_LIMIT = 2000
        job = Gaia.launch_job_async(adql_query)
        result = job.get_results()
    elif gaia_provider.lower() == 'vizier':
        tap = TapPlus(url="https://TAPVizieR.cds.unistra.fr/TAPVizieR/tap")
        job = tap.launch_job(adql_query)
        result_vizier = job.get_results()
        # change column names to what we expect from Gaia:
        result = Table()
        for vizier_col, gaia_col in vizier_to_gaia_column_mapping.items():
            if vizier_col in result_vizier.colnames:
                result[gaia_col] = result_vizier[vizier_col]
        # vizier does not provide the reference epoch.
        result['ref_epoch'] = Column(name='ref_epoch', data=2016.0 * np.ones(len(result), dtype=float))
        # check that we're using dr3, for which the ref epoch is indeed 2016:
        if 'gaiadr3' not in adql_query:
            raise FutureWarning("Using Vizier and 2016 as ref epoch, but not using Gaia DR3. Ref epoch changed? check.")
    else:
        raise RuntimeError("gaia_provider must be 'gaia' or 'vizier'")

    return result


def find_gaia_stars_in_circle(center_radius, gaia_provider='gaia', astrometric_excess_noise_max=None, gmag_range=None,
                              min_phot_g_mean_flux_over_error=None):
    """
    Query Gaia stars within a circle defined by a central point and radius, with additional filtering options.

    :param center_radius: dictionary {'center': tuple(ra,dec), 'radius': radius_degrees}
    :param gaia_provider: str, default 'gaia'. Either 'gaia' or 'vizier'
    :param astrometric_excess_noise_max: float, maximum allowed astrometric excess noise (None for no filter)
    :param gmag_range: tuple, magnitude range in g-band as (min_gmag, max_gmag) (None for no filter)
    :param min_phot_g_mean_flux_over_error: float, flux error correlates to variability. (None for no filter)

    Returns: astropy table of gaia sources
    """
    where_conditions, query_table = construct_where_conditions(gaia_provider, astrometric_excess_noise_max,
                                                               gmag_range, min_phot_g_mean_flux_over_error)

    # Constructing the circle condition
    c = center_radius['center']
    r = center_radius['radius']
    # while handling the vizier weird column naming:
    ra_col = 'ra'
    dec_col = 'dec'
    if gaia_provider == 'vizier':
        ra_col = gaia_to_vizier_column_mapping[ra_col]
        dec_col = gaia_to_vizier_column_mapping[dec_col]
    where_conditions.append(
        f"1=CONTAINS(POINT('ICRS', gdr3.{ra_col}, gdr3.{dec_col}), CIRCLE('ICRS', {c[0]}, {c[1]}, {r}))"
    )

    where_clause = " AND ".join(where_conditions)

    adql_query = f"""
    SELECT * FROM {query_table}
    """
    if where_clause:
        adql_query += f" WHERE {where_clause}"

    result = run_query(gaia_provider=gaia_provider, adql_query=adql_query)
    return result


def find_gaia_stars_in_polygon(vertices, gaia_provider='gaia', astrometric_excess_noise_max=None, gmag_range=None,
                               min_phot_g_mean_flux_over_error=None):
    """
    Query Gaia stars within a polygon defined by a list of vertices, with additional filtering options.

    :param vertices: list of tuples, each tuple contains RA and Dec in degrees [(ra1, dec1), (ra2, dec2), ..., ]
    :param gaia_provider: str, default 'gaia'. Either 'gaia' or 'vizier'
    :param astrometric_excess_noise_max: float, maximum allowed astrometric excess noise (None for no filter)
    :param gmag_range: tuple, magnitude range in g-band as (min_gmag, max_gmag) (None for no filter)
    :param min_phot_g_mean_flux_over_error: float, flux error correlates to variability. (None for no filter)

    Returns: astropy table of gaia sources
    """
    where_conditions, query_table = construct_where_conditions(gaia_provider, astrometric_excess_noise_max,
                                                               gmag_range, min_phot_g_mean_flux_over_error)

    polygon_string = ', '.join([f"{vertex[0]},{vertex[1]}" for vertex in vertices])

    # handle the vizier weird column naming:
    ra_col = 'ra'
    dec_col = 'dec'
    if gaia_provider == 'vizier':
        ra_col = gaia_to_vizier_column_mapping[ra_col]
        dec_col = gaia_to_vizier_column_mapping[dec_col]
    where_conditions.append(
        f"1=CONTAINS(POINT('ICRS', gdr3.{ra_col}, gdr3.{dec_col}), POLYGON('ICRS', {polygon_string}))"
    )

    where_clause = " AND ".join(where_conditions)

    adql_query = f"""
    SELECT * FROM {Gaia.MAIN_GAIA_TABLE}
    WHERE {where_clause}
    """

    result = run_query(gaia_provider=gaia_provider, adql_query=adql_query)
    return result




import numpy as np
import pandas as pd
from scipy.stats import sigmaclip
from copy import deepcopy
import warnings


def group_observations(df, threshold=0.8):
    """
    Groups observations in a DataFrame based on a time threshold.

    Parameters:
    - df: pandas DataFrame containing the observations.
        * Must include a 'mjd' column for Modified Julian Days (floats).
        * Must include pairs of 'fluxes_{ps}' and 'd_fluxes_{ps}' columns for each observed point source,
          where 'ps' is a unique identifier each point source.
        * May include other optional columns like 'seeing', for which the average will be computed in each bin.
    - threshold: float, the time threshold in days to define observation groups. Defaults to 0.8.

    Returns:
    - A new DataFrame containing the combined observations (weighted means with 2-sigmas rejection)
    """

    grouped_results = []
    df_sorted = df.sort_values(by='mjd')
    start_idx = 0

    point_sources = {col.split('_')[0] for col in df.columns if col.endswith('_flux') and not col.endswith('_d_flux')}


    for i in range(1, len(df_sorted)):
        if df_sorted.iloc[i]['mjd'] - df_sorted.iloc[i - 1]['mjd'] > threshold or i == len(df_sorted) - 1:
            end_idx = i if df_sorted.iloc[i]['mjd'] - df_sorted.iloc[i - 1]['mjd'] > threshold else i + 1
            df_group = df_sorted.iloc[start_idx:end_idx]
            avg_mjd = df_group['mjd'].mean()
            scatter_mjd = df_group['mjd'].std()

            this_epoch_group = {"mjd": avg_mjd, "scatter_mjd": scatter_mjd}
            flux_columns = [f'{ps}_flux' for ps in point_sources] + [f'{ps}_d_flux' for ps in point_sources]

            optional_averages = {col: df_group[col].mean() for col in df_group.columns if col not in (['mjd'] + flux_columns )}
            this_epoch_group.update(optional_averages)

            for ps in sorted(point_sources):
                col = f'{ps}_flux'
                d_col = f'{ps}_d_flux'
                curve_data = df_group[col].to_numpy()
                curve_variances = df_group[d_col].to_numpy() ** 2

                filtered_data, low_limit, high_limit = sigmaclip(curve_data, low=2, high=2)
                filtered_indices = np.logical_and(curve_data >= low_limit, curve_data <= high_limit)

                filtered_variances = curve_variances[filtered_indices]
                weights = 1. / filtered_variances
                weighted_mean = np.average(filtered_data, weights=weights)
                weighted_variance = np.average((filtered_data - weighted_mean) ** 2, weights=weights)
                weighted_std_deviation = np.sqrt(weighted_variance)

                this_epoch_group[f'{ps}_flux'] = weighted_mean
                this_epoch_group[f'{ps}_d_flux'] = np.sqrt(1./np.sum(weights))
                this_epoch_group[f'{ps}_scatter_flux'] = weighted_std_deviation
                this_epoch_group[f'{ps}_count_flux'] = len(weights)

            start_idx = end_idx
            grouped_results.append(this_epoch_group)

    return pd.DataFrame(grouped_results)


def convert_flux_to_magnitude(df):
    """
    Converts fluxes and flux errors/scatters (see params for pattern) in a DataFrame to magnitudes,
    assuming a 'zeropoint' column (zero if absent).
    This is very tailored to the output of the `group_observations` function.

    Parameters:
    - df: pandas DataFrame containing fluxes and flux errors.
          The DataFrame should contain columns named "{ps}_flux", "{ps}_d_flux", and "{ps}_flux_scatter"
          for each source {ps}, and optionally a "zeropoint" column.

    Returns:
    - A new pandas DataFrame with magnitudes and magnitude errors/scatters.
    """
    df = deepcopy(df)
    if 'zeropoint' not in df.columns:
        warnings.warn('Zeropoint column missing. Using a zeropoint of 0.', RuntimeWarning)
        df['zeropoint'] = 0
    aux_columns = [c for c in df.columns if '_scatter_flux' in c or '_d_flux' in c or '_count' in c]
    flux_columns = [c for c in df.columns if '_flux' in c and c not in aux_columns]
    scatter_error_columns = [col for col in df.columns if '_scatter_flux' in col]  # can be absent

    for flux_col in flux_columns:
        ps = flux_col.split('_')[0]
        error_col = f'{ps}_d_flux'

        df[f'{ps}_mag'] = -2.5 * np.log10(df[flux_col]) + df['zeropoint']

        df[f'{ps}_d_mag'] = 2.5 / np.log(10) * (df[error_col] / df[flux_col]).abs()

        scatter_col = f'{ps}_scatter_flux'
        if scatter_col in scatter_error_columns:
            df[f'{ps}_scatter_mag'] = 2.5 / np.log(10) * (df[scatter_col] / df[flux_col]).abs()

    return df



import sqlite3
import numpy as np

from ..structure.database import execute_sqlite_query, get_pandas
from ..structure.user_config import get_user_config
from ..utilities.footprint import get_combined_footprint_hash


def convert_gaia_to_photometric_band(gaia_like_pandas_table, band):
    """
    Converts Gaia magnitudes to other photometric magnitudes using polynomial coefficients.
    For example, the sdss coefficients are from table 5.6 of
    https://gea.esac.esa.int/archive/documentation/GEDR3/Data_processing/chap_cu5pho/cu5pho_sec_photSystem/cu5pho_ssec_photRelations.html
    (see, e.g., row g - r, here we do r - g, and add g.)
    Args:
        gaia_like_pandas_table (pd.DataFrame): frame with columns phot_g_mean_mag, phot_rp_mean_mag, phot_bp_mean_mag
        band (str): The desired band:
                'r_sdss', 'i_sdss', 'g_sdss',
                'V', 'R', or 'Ic' for Johnson-Cousins,
                'B_T' or 'V_T' for Tycho2
    Returns:
        pd.Series: A pandas Series containing the specified magnitudes.
    """
    # taken from tables 5.6 and 5.7, see link in docstring
    coefficients = {
        'r_sdss': [-0.09837, 0.08592, 0.1907, -0.1701, 0.02263],
        'i_sdss': [-0.293, 0.6404, -0.09609, -0.002104],
        'g_sdss': [0.2199, -0.6365, -0.1548, 0.0064],
        'V': [-0.02704, 0.01424, -0.2156, 0.01426],
        'R': [-0.02275, 0.3961, -0.1243, -0.01396, 0.003775],
        'Ic': [0.01753, 0.76, -0.0991],
        'V_T': [-0.01077, -0.0682, -0.2387, 0.02342],
        'B_T': [-0.004288, -0.8547, 0.1244, -0.9085, 0.4843, -0.06814]
    }

    if band not in coefficients:
        raise ValueError(f"Unsupported band. Choose among {list(coefficients.keys())}.")

    coef = coefficients[band]
    bp_rp = gaia_like_pandas_table['phot_bp_mean_mag'] - gaia_like_pandas_table['phot_rp_mean_mag']
    g = gaia_like_pandas_table['phot_g_mean_mag']

    mag = g - sum(coef[i] * bp_rp**i for i in range(len(coef)))

    return mag


def calculate_zeropoints():
    """
    for a given the current combined_footprint_hash, queries all the available star fluxes.
    compares them with gaia magnitudes to calculate a magnitude in your band
    (ONLY IMPLEMENTED FOR R-BAND SO FAR)
    TODO
    Then calculates a zeropoint for each frame, and updates the database.
    Args:

    Returns:

    """
    flux_query = """
    SELECT 
         sff.frame_id, 
         sff.flux, 
         s.gaia_id, 
         s.gmag as phot_g_mean_mag, 
         s.bmag as phot_bp_mean_mag,
         s.rmag as phot_rp_mean_mag
    FROM 
         star_flux_in_frame sff
    JOIN 
         stars s ON sff.star_gaia_id = s.gaia_id
    AND
         s.combined_footprint_hash = sff.combined_footprint_hash
    JOIN 
         frames f ON f.id = sff.frame_id
    WHERE 
         sff.combined_footprint_hash = ?
    """
    user_config = get_user_config()

    frames_ini = get_pandas(columns=['id'],
                            conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_ini['id'].to_list())

    flux_data = execute_sqlite_query(flux_query, (combined_footprint_hash,), is_select=True, use_pandas=True)
    if flux_data.empty:
        return

    # so, Gaia magnitudes to other photometric band
    flux_data['converted_mag'] = convert_gaia_to_photometric_band(flux_data, user_config['photometric_band'])

    flux_data['instrumental_mag'] = -2.5 * np.log10(flux_data['flux'])
    flux_data['mag_difference'] = flux_data['converted_mag'] - flux_data['instrumental_mag']

    # zeropoint and its uncertainty (std) for each frame
    zeropoint_results = flux_data.groupby('frame_id')['mag_difference'].agg(['median', 'std']).reset_index()
    zeropoint_results.rename(columns={'median': 'zeropoint', 'std': 'zeropoint_uncertainty'}, inplace=True)

    # and update database
    insert_query = """
    INSERT INTO approximate_zeropoints (frame_id, combined_footprint_hash, zeropoint, zeropoint_uncertainty)
    VALUES (?, ?, ?, ?)
    ON CONFLICT(frame_id, combined_footprint_hash) DO UPDATE SET
    zeropoint = excluded.zeropoint,
    zeropoint_uncertainty = excluded.zeropoint_uncertainty;
    """
    data_to_insert = [(row['frame_id'], combined_footprint_hash, row['zeropoint'], row['zeropoint_uncertainty']) for
                      _, row in zeropoint_results.iterrows()]

    db_path = get_user_config()['database_path']
    with sqlite3.connect(db_path, timeout=15.0) as conn:
        conn.executemany(insert_query, data_to_insert)

